@@ -23,6 +23,8 @@
  *  Copyright (C) 2000-2003 Ximian Inc.
  */
 
+#define GNOME_PRINT_UNSTABLE_API
+
 #include <config.h>
 #include <math.h>
 #include <string.h>
@@ -34,6 +36,7 @@
 #include <libgnomeprint/gpa/gpa-node.h>
 #include <libgnomeprint/gnome-print-private.h>
 #include <libgnomeprint/gnome-print-meta.h>
+#include <libgnomeprint/gnome-print-transport.h>
 #include <libgnomeprint/gnome-print-multipage.h>
 #include <libgnomeprint/gnome-print-job-private.h>
 #include <libgnomeprint/gnome-print-config-private.h>
@@ -62,6 +65,13 @@ struct _JOBPrivate {
 	GList *LY_LIST;
 };
 
+enum
+{
+  PROP_0,
+  PROP_CONFIG,
+};
+ 
+
 #define GNOME_PRINT_JOB_CLOSED(m) (((JOBPrivate *) (m)->priv)->closed)
 
 static void gnome_print_job_class_init (GnomePrintJobClass *klass);
@@ -73,8 +83,18 @@ static void job_parse_config_data (GnomePrintJob *job);
 static void job_clear_config_data (GnomePrintJob *job);
 static gboolean job_parse_transform (guchar *str, gdouble *transform);
 
+static void gnome_print_job_get_property (GObject      *object,
+                                          guint         prop_id,
+                                          GValue       *value,
+                                          GParamSpec   *pspec);
+static void gnome_print_job_set_property (GObject      *object,
+                                          guint         prop_id,
+                                          const GValue *value,
+                                          GParamSpec   *pspec);
+
 static GObjectClass *parent_class;
 
+
 GType
 gnome_print_job_get_type (void)
 {
@@ -100,6 +120,18 @@ gnome_print_job_class_init (GnomePrintJobClass *klass)
 	GObjectClass *object_class;
 	
 	object_class = (GObjectClass *) klass;
+
+	object_class->set_property = gnome_print_job_set_property;
+	object_class->get_property = gnome_print_job_get_property;
+
+	g_object_class_install_property (object_class,
+                                   PROP_CONFIG,
+                                   g_param_spec_object ("config",
+                                                        "Job Configuration",
+                                                        "The configuration for the print job",
+                                                        GNOME_TYPE_PRINT_CONFIG,
+                                                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+  
 	parent_class = g_type_class_peek_parent (klass);
 
 	object_class->finalize = gnome_print_job_finalize;
@@ -110,7 +142,9 @@ gnome_print_job_init (GnomePrintJob *job)
 {
 	job->config = NULL;
 
-	job->meta = gnome_print_meta_new ();
+	job->meta = NULL;
+
+	job->input_file = NULL;
 
 	job->priv = g_new0 (JOBPrivate, 1);
 	job_clear_config_data (job);
@@ -125,12 +159,17 @@ gnome_print_job_finalize (GObject *object)
 
 	if (job->config) {
 		g_object_unref (G_OBJECT (job->config));
+		job->config = NULL;
 	}
 
 	if (job->meta != NULL) {
 		g_object_unref (G_OBJECT (job->meta));
+		job->meta = NULL;
 	}
 
+	g_free (job->input_file);
+	job->input_file = NULL;
+
 	if (job->priv) {
 		job_clear_config_data (job);
 		g_free (job->priv);
@@ -140,6 +179,55 @@ gnome_print_job_finalize (GObject *object)
 	G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
+static void
+gnome_print_job_get_property (GObject *object, guint prop_id, GValue *value,
+	                     GParamSpec *pspec)
+{
+  GnomePrintJob *job = GNOME_PRINT_JOB (object);
+
+  switch (prop_id)
+    {
+    case PROP_CONFIG:
+      g_value_set_object (value, gnome_print_job_get_config (job));
+      break;
+  
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      break;
+    }
+}
+
+static void
+gnome_print_job_set_property (GObject * object, guint prop_id,
+	                     const GValue * value, GParamSpec * pspec)
+{
+  GnomePrintJob *job = GNOME_PRINT_JOB (object);
+  GnomePrintConfig *config = NULL;
+
+  switch (prop_id)
+    {
+    case PROP_CONFIG:
+	    config = GNOME_PRINT_CONFIG (g_value_get_object (value));
+	    if (job->config) {
+		    gnome_print_config_unref (job->config);
+		    job->config = NULL;
+	    }
+	    
+	    if(config) {
+		    gnome_print_config_ref (config);
+		    job->config = config;
+	    } else {
+		    /* Use the default: */
+		    job->config = gnome_print_config_default ();
+	    }
+	    break;
+    default:
+	    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+	    break;
+    }
+}
+
+
 /**
  * gnome_print_job_new:
  * @config: The job options, can be NULL in which case a
@@ -154,12 +242,7 @@ gnome_print_job_new (GnomePrintConfig *config)
 {
 	GnomePrintJob *job;
 	
-	job = g_object_new (GNOME_TYPE_PRINT_JOB, NULL);
-
-	if (config)
-		job->config = gnome_print_config_ref (config);
-	else
-		job->config = gnome_print_config_default ();
+	job = g_object_new (GNOME_TYPE_PRINT_JOB, "config", config, NULL);
 
 	return job;
 }
@@ -198,13 +281,39 @@ gnome_print_job_get_context (GnomePrintJob *job)
 {
 	g_return_val_if_fail (job != NULL, NULL);
 	g_return_val_if_fail (GNOME_IS_PRINT_JOB (job), NULL);
+	g_return_val_if_fail (job->input_file == NULL, NULL);
 
 	if (job->meta)
-		g_object_ref (G_OBJECT (job->meta));
+	g_object_ref (G_OBJECT (job->meta));
+	else
+		job->meta = gnome_print_meta_new ();
 
 	return job->meta;
 }
 
+/**
+ * gnome_print_job_set_file:
+ * @job: An initialised GnomePrintJob.
+ * @input_file: The file to input from.
+ * 
+ * Set the file to print from. This allows applications to use
+ * gnome-print without using the drawing API by providing the
+ * postscipt output.
+ * 
+ * Return value: None
+ **/
+void
+gnome_print_job_set_file (GnomePrintJob *job, gchar *input_file)
+{
+	g_return_if_fail (job != NULL);
+	g_return_if_fail (GNOME_IS_PRINT_JOB (job));
+	g_return_if_fail (input_file != NULL);
+	g_return_if_fail (job->input_file == NULL);
+	g_return_if_fail (job->meta == NULL);
+
+	job->input_file = g_strdup (input_file);
+}
+
 /**
  * gnome_print_job_get_pages:
  * @job: An initialised and closed GnomePrintJob.
@@ -224,9 +333,8 @@ gnome_print_job_get_pages (GnomePrintJob *job)
 
 	g_return_val_if_fail (job != NULL, 0);
 	g_return_val_if_fail (GNOME_IS_PRINT_JOB (job), 0);
-
-	if (!GNOME_PRINT_JOB_CLOSED (job))
-		return 0;
+	g_return_val_if_fail (job->meta, 0);
+	g_return_val_if_fail (GNOME_PRINT_JOB_CLOSED (job), 0);
 
 	job_update_layout_data (job);
 
@@ -313,6 +421,10 @@ gnome_print_job_close (GnomePrintJob *job)
 
 	g_return_val_if_fail (job != NULL, GNOME_PRINT_ERROR_UNKNOWN);
 	g_return_val_if_fail (GNOME_IS_PRINT_JOB (job), GNOME_PRINT_ERROR_UNKNOWN);
+	g_return_val_if_fail (job->input_file == NULL, 0);
+
+	if (!job->meta)
+		job->meta = gnome_print_meta_new ();
 
 	pp = job->priv;
 
@@ -326,6 +438,69 @@ gnome_print_job_close (GnomePrintJob *job)
 	return GNOME_PRINT_ERROR_UNKNOWN;
 }
 
+/**
+ * gnome_print_job_metadata_printer:
+ * @job: 
+ * @retval: 
+ * 
+ * We have to special case the Metadata printer where we want to create a file on
+ * disk with the drawing commands to later play them. libgnomeprint needs to be
+ * configured with --with-metadata-printer for it to show up in the printers list
+ * this is used to create streams of drawing commands for example for the regression
+ * test suite.
+ * 
+ * Return Value: TRUE if this job is using the Metadata (to file) printer
+ **/
+static gboolean
+gnome_print_job_metadata_printer (GnomePrintJob *job, gint *retval)
+{
+	GnomePrintTransport *transport;
+	gchar *drivername;
+	gboolean print_to_file = FALSE;
+	const guchar *buf;
+	gint blen;
+
+	drivername = gnome_print_config_get (job->config, "Settings.Engine.Backend.Driver");
+
+	if (drivername == NULL)
+		return FALSE;
+
+	if (strcmp (drivername, "gnome-print-meta") != 0) {
+		g_free (drivername);
+		return FALSE;
+	}
+
+	*retval = GNOME_PRINT_ERROR_UNKNOWN;
+	
+	gnome_print_config_get_boolean (job->config, "Settings.Output.Job.PrintToFile", &print_to_file);
+	if (!print_to_file) {
+		g_warning ("Metadata printer should always be print to file");
+		goto metadata_printer_done;
+	}
+
+	transport = gnome_print_transport_new (job->config);
+	if (!transport) {
+		g_warning ("Could not create transport for metadata printer");
+		goto metadata_printer_done;
+	}
+
+	buf  = gnome_print_meta_get_buffer (GNOME_PRINT_META (job->meta));
+	blen = gnome_print_meta_get_length (GNOME_PRINT_META (job->meta));
+
+	gnome_print_transport_open  (transport);
+	gnome_print_transport_write (transport, buf, blen);
+	gnome_print_transport_write (transport, "GNOME_METAFILE_END", strlen ("GNOME_METAFILE_END"));
+	gnome_print_transport_close (transport);
+	
+	*retval = GNOME_PRINT_OK;
+
+metadata_printer_done:
+
+	g_free (drivername);
+
+	return TRUE;
+}
+
 /**
  * gnome_print_job_print:
  * @job: A closed GnomePrintJob.
@@ -351,6 +526,11 @@ gnome_print_job_print (GnomePrintJob *job)
 	g_return_val_if_fail (GNOME_IS_PRINT_JOB (job), GNOME_PRINT_ERROR_UNKNOWN);
 	g_return_val_if_fail (job->priv, GNOME_PRINT_ERROR_UNKNOWN);
 
+	if (job->input_file) {
+		GnomePrintTransport *transport = gnome_print_transport_new (job->config);
+		return gnome_print_transport_print_file (transport, job->input_file);
+	}
+
 	pp = job->priv;
 
 	if (!pp->closed) {
@@ -358,9 +538,6 @@ gnome_print_job_print (GnomePrintJob *job)
 			   "gnome_print_job_print\n");
 		gnome_print_job_close (job);
 	}
-	
-	ctx = gnome_print_context_new (job->config);
-	g_return_val_if_fail (ctx != NULL, GNOME_PRINT_ERROR_UNKNOWN);
 
 	/* Get number of pages in metafile */
 	lpages = gnome_print_meta_get_pages (GNOME_PRINT_META (job->meta));
@@ -368,7 +545,15 @@ gnome_print_job_print (GnomePrintJob *job)
 		return GNOME_PRINT_OK;
 	npages = lpages;
 
-	/* Update layout data */
+	/* Special case where printer is a Metadata print to file */
+	if (gnome_print_job_metadata_printer (job, &ret))
+		return ret;
+
+	/* Create the "real" context */
+	ctx = gnome_print_context_new (job->config);
+	g_return_val_if_fail (ctx != NULL, GNOME_PRINT_ERROR_UNKNOWN);
+
+	/* Update the layout data */
 	is_multipage = FALSE;
 	job_update_layout_data (job);
 	if (pp->LY_LIST) {
@@ -448,6 +633,7 @@ gnome_print_job_render (GnomePrintJob *job, GnomePrintContext *ctx)
 	g_return_val_if_fail (ctx != NULL, GNOME_PRINT_ERROR_UNKNOWN);
 	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (ctx), GNOME_PRINT_ERROR_UNKNOWN);
 	g_return_val_if_fail (job->priv, GNOME_PRINT_ERROR_UNKNOWN);
+	g_return_val_if_fail (job->meta, GNOME_PRINT_ERROR_UNKNOWN);
 
 	pp = job->priv;
 
@@ -494,6 +680,7 @@ gnome_print_job_render_page (GnomePrintJob *job, GnomePrintContext *ctx, gint pa
 	g_return_val_if_fail (ctx != NULL, GNOME_PRINT_ERROR_UNKNOWN);
 	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (ctx), GNOME_PRINT_ERROR_UNKNOWN);
 	g_return_val_if_fail (job->priv, GNOME_PRINT_ERROR_UNKNOWN);
+	g_return_val_if_fail (job->meta, GNOME_PRINT_ERROR_UNKNOWN);
 
 	pp = job->priv;
 
@@ -722,17 +909,22 @@ job_parse_config_data (GnomePrintJob *job)
 				gint pagenum;
 				affines = g_new (gdouble, 6 * numlp);
 				pagenum = 0;
-				for (page = gpa_node_get_child (pages, NULL); page != NULL; page = gpa_node_get_child (pages, page)) {
+
+				while (pagenum < numlp) {
 					guchar *transform;
+					guchar *child_id;
+					child_id = g_strdup_printf ("LP%d", pagenum);
+					page = gpa_node_get_child_from_path (pages, child_id);
 					transform = gpa_node_get_value (page);
 					gpa_node_unref (page);
-					if (!transform)
+					if (!transform) {
+						g_warning ("Could not fetch transfrom from %s\n", child_id);
 						break;
+					}
 					job_parse_transform (transform, affines + 6 * pagenum);
 					g_free (transform);
+					g_free (child_id);
 					pagenum += 1;
-					if (pagenum >= numlp)
-						break;
 				}
 				gpa_node_unref (pages);
 				if (pagenum == numlp) {
@@ -1114,9 +1306,8 @@ gnome_print_job_print_to_file (GnomePrintJob *job, gchar *output)
 		gnome_print_config_set (job->config, "Settings.Transport.Backend",    "file");
 		gnome_print_config_set (job->config, GNOME_PRINT_KEY_OUTPUT_FILENAME, output);
 	} else {
-		g_print ("Setting print to file to off, using LPR\n");
 		/* In the future we might want to use the default or even better,
-		 * go back to the prev. selected printer
+		 * go back to the prev. selected printer (Chema)
 		 */
 		gnome_print_config_set (job->config, "Settings.Transport.Backend", "lpr");
 	}
@@ -1141,7 +1332,6 @@ gnome_print_config_get_transform (GnomePrintConfig *config, const guchar *key, g
 
 	v = gnome_print_config_get (config, key);
 	if (!v) {
-		g_print ("Could not get %s\n", key);
 		return FALSE;
 	}
 
